' A Very Small Fast Parser using PB Inline Assembler

' including demo with performance test

' Charles E V Pegge
' 3 August 2006
' charles@pegge.net
' www.pegge.net


' this parser is suitable for use in a script engine or for parsing any data strings
' that use symbols as well as words. Call it repeatedly to indicate sequential words
' and symbols in a string of text.

' I have written it to accelerate my scripting language; the machine code is only 251 bytes
' long and goes like a rocket! On the test my PC pushes out about 19 million words per second.

' General parsing rules:
'
' space, and all lower ascii and all non alphanumerics are treated as delimiters. eg:
' But an underscore _ (ascii 95) is treated as an alphanumeric eg:
'
'   FREE.LUNCH is interpreted as three words
'   FREE_LUNCH is one word

' all other non alphanumerics including '-' (minus) are self delimiting eg:
'
'   + - . ,  { } are always recognised as individual words even when unspaced
'
'
' But quoted text is treated as an entire word including the outer quotes.
' inner quotes can be embedded eg:
'
'    "She said 'HI' "
'    'She said "HI" '
'
' this code uses the EAX ECX EDX and EBP registers and holds 2 dwords on the stack -
' with minimal use of memory, most of the work is done in registers.
' higher speeds are possible by loading string in 4byte dwords instead of individual bytes
' but with more complex code.
' Another possibility is using short jumps instead of long jumps - but these are not supported
' by the PB assembler so the binary would have to be manually tweaked for this.
' The length of the code would be reduced from about 250 bytes to 190 bytes, significantly increasing
' execution speedd.
'
' you can use it inline or in its own subroutine or function.
'


' quick Ascii code reference:

'  various control chars 0-31:
'  null      0
'  htab      9
'  linefeed  10
'  vtab      11
'  formfeed  12
'  creturn   13
'  escape    27
'  space     32
'  !"#$%&'   33=39
'  ()*+,-./  40-47
'  0-9       48-57
'  :;<=>?@   58-64
'  A-Z       65-90
'  [\]^_`    91-96
'  a-z       97-122
'  {|}~      123-126



#COMPILE EXE
#DIM ALL
#INCLUDE "WIN32API.INC"

FUNCTION PBMAIN () AS LONG

' input/output variables

DIM az  AS STRING   ' Test string
DIM bwi AS DWORD    ' character position of word in the string
DIM lbi AS DWORD    ' length of string
DIM ii  AS DWORD    ' next position after the end of the word
DIM ip  AS BYTE PTR ' absolute address of the string

' test variables

DIM result    AS STRING
DIM testloops AS DWORD
DIM timing    AS DWORD
DIM wordcount AS DWORD

' test string

az="Old_wine    Free.Meal(Lunch,12.34)  "+CHR$(34)+"bon mot (7)"+CHR$(34)+"  "+CHR$(39)+CHR$(34)+"okay"+CHR$(34)+CHR$(39)

'measure speed

lbi=LEN(az)
ip=STRPTR(az)-1
timing=GetTickCount() ' ticks every millisecond
FOR testloops=1 TO 1000000 ' 1 million
 ii=1
 WHILE ii=<lbi
  GOSUB zbword
 WEND
NEXT
timing=GetTickCount()-Timing


'show results

result=az+":"+$CRLF
ii=1
WHILE ii=<lbi
 INCR wordcount
 GOSUB zbword
 result=result+MID$(az,bwi,ii-bwi)+$CRLF
WEND
MSGBOX result+$CRLF _
 +"words/millisec "+STR$(INT(1000000*wordcount/timing))+$CRLF _
 +"Length of code:"+STR$(CODEPTR(zeword)-CODEPTR(zbword))

EXIT FUNCTION

' the code itself

zbword:

! push ebp                  ' save original base pointer
! mov edx,ip                ' pointer to start of string-1
! push edx                  ' hold this pointer on the stack
! mov eax,edx               ' to find end of string
! add edx,lbi               ' points to the last character in the string
! add eax,ii                ' points to the character indexed by ii
! mov ecx,eax               ' default value for the start of word pointer
! dec eax                   ' prepare for pre incrementing loop
                            ' with basic variables installed,
                            ' it is now safe to overwrite the base pointer register
! mov ebp,eax               ' move the char pointer to the base pointer
! xor eax,eax               ' clear eax for another use


next_space:                 ' for skipping leading spaces tabs etc
! inc ebp                   ' advance the char pointer
! cmp edx,ebp               ' check if end of string
! jl start_of_word          ' exit loop if end of string
! mov al,[ebp]              ' get character
! cmp al,32                 ' is it a space or less ?
! jle next_space            ' if so then try text character

start_of_word:              ' first non space char encountered

! mov ecx,ebp               ' remember start position

! cmp edx,ebp               ' check if end of string
! jl end_of                 ' exit if end of string
! cmp al,34                 ' double quote ?
! jz quote_string           ' process whole quote
! cmp al,39                 ' single quote ?
! jz quote_string           ' process whole quote

! cmp al,48                 ' is it less than a number?
! jl self_delimited         ' must be a symbol
! cmp al,122                ' is it greater than lowercase z ?
! jg self_delimited         ' must be a symbol or some higher ascii code
! cmp al,58                 ' is it a number ie 48 to 57
! jl next_letter            ' if so then look at next char
! cmp al,65                 ' it it less than uppercase A ?
! jl self_delimited         ' if so then it must be a symbol
! cmp al,91                 ' is it capital Z or less ?
! jl next_letter            ' if so then it must be an upper case alpha char, try next
! cmp al,95                 ' is it an underscore ?
! jz next_letter            ' then treat as a letter
! cmp al,97                 ' is it less than lowercase a?
! jl self_delimited         ' if so then it must be a symbol
! jmp next_letter           ' we are left with lowercase alphas, try next

self_delimited:

! inc ebp                   ' increment then char pointer beyond the end of the word
! jmp end_of                ' the word is a single char symbol, finish parsing

quote_string:               ' this is a quoted item, we need to find the end quote mark

quote_string_traverse:      ' look for a matching quote mark

! inc ebp                   ' advance pointer for next char
! cmp edx,ebp               ' check if past the end of string
! jl end_of                 ' finish if so
! cmp al,[ebp]              ' check for matching quote mark
! jnz quote_string_traverse ' if no match then try next
! inc ebp                   ' advance past end of quote
! jmp end_of                ' we have our quote so finish

next_letter:                ' scanning an alpha numeric. Could be delimited by a symbol, space or lesser ascii

! inc ebp                   ' advance pointer
! cmp edx,ebp               ' past end of string?
! jl end_of                 ' if so then finish
! mov al,[ebp]              ' look at char
! cmp al,48                 ' if its less than a number
! jl end_of                 ' then finish
! cmp al,122                ' if its greater than lowercase z ?
! jg end_of                 ' then finish
! cmp al,58                 ' is it a number ?
! jl next_letter            ' then continue
! cmp al,65                 ' is it less than uppercase A
! jl end_of                 ' then its a symbol delimiting the word so finish
! cmp al,91                 ' is it an uppercase alpha?
! jl next_letter            ' then continue
! cmp al,95                 ' is it an underscore ?
! jz next_letter            ' then treat as a letter
! cmp al,97                 ' is it less than lowercase a ?
! jl end_of                 ' then it is a symbol so finish
! jmp next_letter           ' must be lowercase alpha so continue

end_of:                     ' results are passed back to BASIC variables

! pop edx                   ' original position of pointer
! mov eax,ebp               ' end position
! pop ebp                   ' restore base pointer so it is safe to assign to BASIC variables
! sub ecx,edx               ' derive our start of word index
! mov bwi,ecx               ' store result in bwi
! sub eax,edx ; derive ii   ' derive the after-word index
! mov ii,eax                ' assign value to ii

                            ' if the string is null or contains no words or symbols
                            ' then both bwi and ii will be greater than the length of the string

zeword:                    ' this label is just for measuring the length of the assembled code

RETURN

END FUNCTION
